#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
/* Save all file descriptor flags in an array */
int* c_save_fdflags(void)
{
  int  maxfds = sysconf(_SC_OPEN_MAX);
  int* flags  = calloc(maxfds, sizeof(int));
  int  fd;

  for (fd = 0; fd < maxfds; fd++)
     /* Saves -1 for invalid fds */
     flags[fd] = fcntl(fd, F_GETFL);

  return flags;
}

/* Restore all file descriptor flags from the array, and free it */
void c_restore_fdflags(int* flags)
{
  int  maxfds = sysconf(_SC_OPEN_MAX);
  int  fd;

  for (fd = 0; fd < maxfds; fd++)
     if (flags[fd] != -1)
        fcntl(fd, F_SETFL, flags[fd]);

  free(flags);
}

/* Duplicate a file descriptor, allocating the new one at min or above */
int c_fcntl_dupfd(int fd, int min)
{
  return fcntl(fd, F_DUPFD, min);
}

/* Prepare all file descriptors for a subsequent exec */
void c_prepare_fd_flags_for_exec(void)
{
  int maxfds = sysconf(_SC_OPEN_MAX);
  int fd, flags;

  /* Set fds 0-2 to blocking mode */
  for (fd = 0; fd < 3; fd++) {
     flags = fcntl(fd, F_GETFL);
     fcntl(fd, F_SETFL, flags & ~O_NONBLOCK);
  }

  /* Set all other fds to close-on-exec */
  for (fd = 3; fd < maxfds; fd++) {
     flags = fcntl(fd, F_GETFL);
     fcntl(fd, F_SETFL, flags | FD_CLOEXEC);
  }
}




