-- GENERATED by C->Haskell Compiler, version 0.13.4 "Pressing Forward", 17 Oct 2004 (Haskell)
-- Edit the ORIGNAL .chs file instead!


{-# LINE 1 "src/lib/HsShellScript/ProcErr.chs" #-}-- #hide
module HsShellScript.ProcErr where


import Data.Dynamic
import Control.Exception
import Foreign
import Foreign.C
import Foreign.C.Error
import HsShellScript.Args
import HsShellScript.Shell
import IO hiding (bracket)
import Int
import Maybe
import System.Posix
import System
import qualified Control.Exception
import GHC.IOBase
import GHC.Handle
import GHC.Conc                         -- for takeMVar
import Monad
import List

infixr 2 -|-    -- left handed, stdout
infixr 2 =|-    -- left handed, stderr
infixl 2 -|=    -- right handed, stdout
infixl 2 =|=    -- right handed, stderr
infixl 3 ->-    -- write stdout to file
infixl 3 =>-    -- write stderr to file
infixl 3 ->>-   -- append stdout to file
infixl 3 =>>-   -- append stderr to file
infixl 3 -<-    -- read stdin from file or string
infixl 3 -&>-   -- write stdout and stderr to file
infixl 3 -&>>-  -- append stdout and stderr to file


{- | Improved version of @System.Posix.Files.setFileMode@, which sets the file name in the @IOError@ which is thrown in case of an error. The
   implementation in GHC 6.2.2 neglects to do this.

>setFileMode' path mode =
>   fill_in_filename path $
>      setFileMode path mode
-}
setFileMode' :: FilePath -> FileMode -> IO ()
setFileMode' path mode =
   fill_in_filename path $
      setFileMode path mode

-- openFd, with file name filled in
openFd' a b c d = do
   System.Posix.openFd a b c d
      `IO.catch` (\ioe -> ioError (ioe { ioe_filename = Just a }))


-- |
-- Execute an IO action as a separate process, and wait for it to finish.
-- Report errors as exceptions.
--
-- The program forks a child process and performs the specified action.
-- Then it waits for the child process to finish. If it exits in any way
-- which indicates an error, the @ProcessStatus@ is thrown as a dynamic
-- exception.
--
-- The parent process waits for the child process, if it has been stopped with the @STOP@ signal.
--
-- See "HsShellScript#subr" for further details.
--
-- See 'spawn'.
call :: IO a  -- ^ action to execute as a child process
     -> IO ()
call io = do
    pid <- spawn_loc "call" io
    (Just ps) <- getProcessStatus True False pid
    if ps == Exited ExitSuccess
        then return ()
        else Control.Exception.throwDyn ps


-- |
-- Execute an IO action as a separate process, and continue without waiting
-- for it to finish.
--
-- The program forks a child process, which performs the specified action and terminates.
-- The child's process ID is returned.
--
-- See "HsShellScript#subr" for further details.
--
-- See 'call'.
spawn :: IO a           -- ^ Action to execute as a child process.
      -> IO ProcessID   -- ^ Process ID of the new process.
spawn = spawn_loc "spawn"

spawn_loc :: String -> IO a -> IO ProcessID
spawn_loc loc io = do
   flush_outerr
   pid <- fork
{-# LINE 102 "src/lib/HsShellScript/ProcErr.chs" #-}
   case pid of
      0  -> child io
      -1 -> throwErrno' loc Nothing Nothing
      _  -> return (fromIntegral pid)


-- |
-- Run an external program. This starts an external program as a child
-- process, and waits for it to finish. The executable is searched via the
-- @PATH@.
--
-- When the specified program can't be executed, an error message is printed, and the main process
-- gets a @ProcessStatus@ thrown as a dynamic exception, with the value @Exited
-- (ExitFailure 1)@. This means that the main program can't distinguish between
-- failure of calling the program and the program exiting with an exit code of
-- 1.
--
-- @run prog par@ is essentially @call (execp prog par)@.
--
-- Example:
--
-- >run "/usr/bin/foobar" ["some", "args"]
-- >   `catchDyn` (\ps -> do -- oops...
-- >              )
--
-- See 'call', 'spawn'.
run :: FilePath                    -- ^ Name of the executable to run
    -> [String]                    -- ^ Command line arguments
    -> IO ()
run prog par =
   call (execp prog par
            `IO.catch` (\ioe -> do errm ("Error calling " ++ shell_command prog par ++ ":\n" ++ ioe_description ioe)
                                   _exit 1
                       )
        )


-- | Print an action as a shell command, then perform it.
--
-- This is used with actions such as 'run', 'exec' or 'call'. For instance,
-- @echo run prog args@ is a variant of @run prog args@, which prints what
-- is being done before doing it.
--
-- See 'run', 'call', 'exec'.
echo :: ( FilePath -> [String] -> IO () )       -- ^ Action to perform
     -> FilePath                                -- ^ Name or path of the executable to run
     -> [String]                                -- ^ Command line arguments
     -> IO ()
echo action path args = do
   putStrLn (shell_command path args)
   action path args


-- | Execute an external program. This replaces the running process. The path isn't searched, the environment isn't changed. In case of failure,
-- an IOError is thrown.
--
-- >exec path args =
-- >   execute_file path False args Nothing
--
-- See 'execute_file', "HsShellScript#exec".
exec :: String          -- ^ Full path to the executable
     -> [String]        -- ^ Command line arguments
     -> IO a            -- ^ Never returns
exec path args =
   execute_file path False args Nothing


-- | Execute an external program. This replaces the running process. The path is searched, the environment isn't changed. In case of failure,
-- an IOError is thrown.
--
-- >execp prog args =
-- >   execute_file prog True args Nothing
--
-- See 'execute_file', "HsShellScript#exec".
execp :: String        -- ^ Name or path of the executable
      -> [String]      -- ^ Command line arguments
      -> IO a          -- ^ Never returns
execp prog args =
   execute_file prog True args Nothing


-- | Execute an external program. This replaces the running process. The path isn't searched, the environment of the program is set as specified. In
-- case of failure, an IOError is thrown.
--
-- >exece path args env =
-- >   execute_file path False args (Just env)
--
-- See 'execute_file', "HsShellScript#exec".
exece :: String                 -- ^ Full path to the executable
      -> [String]               -- ^ Command line arguments
      -> [(String,String)]      -- ^ New environment
      -> IO a                   -- ^ Never returns
exece path args env =
   execute_file path False args (Just env)


-- | Execute an external program. This replaces the running process. The path is searched, the environment of the program is set as specified. In
-- case of failure, an IOError is thrown.
--
-- >execpe prog args env =
-- >   execute_file prog True args (Just env)
--
-- See 'execute_file', "HsShellScript#exec".
execpe :: String                -- ^ Name or path of the executable
       -> [String]              -- ^ Command line arguments
       -> [(String,String)]     -- ^ New environment
       -> IO a                  -- ^ Never returns
execpe prog args env =
   execute_file prog True args (Just env)


{- | Build left handed pipe of stdout.

   \"@p -|- q@\" builds an IO action from the two IO actions @p@ and @q@.
   @q@ is executed in an external process. The standard output of @p@ is sent
   to the standard input of @q@ through a pipe. The result action consists
   of forking off @q@ (connected with a pipe), and @p@.

   The result action does /not/ run @p@ in a separate process. So, the pipe
   itself can be seen as a modified action @p@, forking a connected @q@.
   Normally, the pipe itself will be forked, too. The pipe is called \"left
   handed\", because @p@ has this property, and not @q@.

   /The exit code of q is silently ignored./ The process ID of the forked
   copy of @q@ isn't returned to the caller, so it's lost.

   See "HsShellScript#subr" and
   "HsShellScript#exec" for further details.

   Examples:

>call (exec "/usr/bin/foo" [] -|- exec "/usr/bin/bar" [])

>call (    execp "foo" ["..."]
>      -|= ( -- Do something with foo's output
>            do cnt <- lazy_contents "-"
>               ...
>          )
>     )

   See 'call', '(=|-)', '(-|=)'.
-}
(-|-) :: IO a   -- ^ Action which won't be forked
      -> IO b   -- ^ Action which will be forked and connected with a pipe
      -> IO a   -- ^ Result action
p -|- q = do
   (Just h, _, _, _) <- pipe_fork_dup q True False False
   hDuplicateTo h stdout
   hClose h
   p


{- | Build left handed pipe of stderr.

   \"@p =|- q@\" builds an IO action from the two IO actions @p@ and @q@.
   @q@ is executed in an external process. The standard error output of @p@ is sent
   to the standard input of @q@ through a pipe. The result action consists
   of forking off @q@ (connected with a pipe), and @p@.

   The result action does /not/ run @p@ in a separate process. So, the pipe
   itself can be seen as a modified action @p@, forking a connected @q@.
   Normally, the pipe itself will be forked, too. The pipe is called \"left
   handed\", because @p@ has this property, and not @q@.

   /The exit code of q is silently ignored./ The process ID of the forked
   copy of @q@ isn't returned to the caller, so it's lost.

   See "HsShellScript#subr" and
   "HsShellScript#exec" for further details.

   Example:

>call (exec "/usr/bin/foo" [] =|- exec "/usr/bin/bar" [])

   See 'call', '(-|-)', '(-|=)'.
-}
(=|-) :: IO a    -- ^ Action which won't be forked
      -> IO b    -- ^ Action which will be forked and connected with a pipe
      -> IO a    -- ^ Result action
p =|- q = do
   (Just h, _, _, _) <- pipe_fork_dup q True False False
   hDuplicateTo h stderr
   hClose h
   p


{- | Build right handed pipe of stdout.

\"@p -|= q@\" builds an IO action from the two IO actions @p@ and @q@.
@p@ is executed in an external process. The standard output of @p@ is sent
to the standard input of @q@ through a pipe. The result action consists
of forking off @p@ (connected with a pipe), and @q@.

The result action does /not/ run @q@ in a separate process. So, the pipe
itself can be seen as a modified action @q@, forking a connected @p@.
Normally, the pipe itself will be forked, too. The pipe is called \"right
handed\", because @q@ has this property, and not @p@.

/The exit code of p is silently ignored./ The process ID of the forked
copy of @q@ isn't returned to the caller, so it's lost.

See "HsShellScript#subr" and
"HsShellScript#exec" for further details.

Example:

>@call (exec \"\/usr\/bin\/foo\" [] -|= exec \"\/usr\/bin\/bar\" [])@

See 'call', '(=|-)', '(=|=)'.
-}
(-|=) :: IO a     -- ^ Action which will be forked and connected with a pipe
      -> IO b     -- ^ Action which won't be forked
      -> IO b     -- ^ Result action
p -|= q = do
   (_, Just h, _, _) <- pipe_fork_dup p False True False
   hDuplicateTo h stdin
   hClose h
   q

{- | Build right handed pipe of stderr.

\"@p =|= q@\" builds an IO action from the two IO actions @p@ and @q@.
@p@ is executed in an external process. The standard error output of @p@ is sent
to the standard input of @q@ through a pipe. The result action consists
of forking off @p@ (connected with a pipe), and @q@.

The result action does /not/ run @q@ in a separate process. So, the pipe
itself can be seen as a modified action @q@, forking a connected @p@.
Normally, the pipe itself will be forked, too. The pipe is called \"right
handed\", because @q@ has this property, and not @p@.

/The exit code of p is silently ignored./ The process ID of the forked
copy of @q@ isn't returned to the caller, so it's lost.

See "HsShellScript#subr" and
"HsShellScript#exec" for further details.

Example:

> call (exec "/usr/bin/foo" [] =|= exec "/usr/bin/bar" [])

See 'call', '=|-', '-|='.
-}
(=|=) :: IO a     -- ^ Action which will be forked and connected with a pipe
      -> IO b     -- ^ Action which won't be forked
      -> IO b     -- ^ Result action
p =|= q = do
   (_, _, Just h, _) <- pipe_fork_dup p False False True
   hDuplicateTo h stdin
   hClose h
   q


redirect_helper stdh mode io path =
   bracket (do bak <- hDuplicate stdh
               h <- openFile path mode
               hDuplicateTo h stdh
               hClose h
               return bak
           )
           (\bak -> do hDuplicateTo bak stdh
                       hClose bak
           )
           (const io)


{- | Redirect the standard output of the specified IO action to a file. The file will be overwritten, if it already exists.

What's actually modified is the @stdout@ handle, not the file descriptor 1. The
@exec@ functions know about this. See "HsShellScript#fdpipes" and
"HsShellScript#exec" for details.

Example:

>call (execp "foo" [] ->- "bar")

See 'call', 'run', '->>-', '=>-'.
-}
(->-) :: IO a           -- ^ Action, whose output will be redirected
      -> FilePath       -- ^ File to redirect the output to
      -> IO a           -- ^ Result action
(->-) =
   redirect_helper stdout WriteMode


{- | Redirect the standard output of the specified IO action to a file. If the file already exists, the output will be appended.

What's actually modified is the @stdout@ handle, not the file descriptor 1. The
@exec@ functions know about this. See "HsShellScript#fdpipes" and
"HsShellScript#exec" for details.

Example:

@call (exec \"\/path\/to\/foo\" [] -\>\>- \"bar\")@

See 'call', 'run', '(->-)', '(=>>-)'.
-}
(->>-) :: IO a          -- ^ Action, whose output will be redirected
       -> FilePath      -- ^ File to redirect the output to
       -> IO a          -- ^ Result action
(->>-) =
   redirect_helper stdout AppendMode


{- | Redirect the standard error output of the specified IO action to a file. If the file already exists, it will be overwritten.

What's actually modified is the @stderr@ handle, not the file descriptor 2. The
@exec@ functions know about this. See "HsShellScript#fdpipes" and
"HsShellScript#exec" for details.

Example: @call (exec \"\/path\/to\/foo\" [] =\>- \"\/dev\/null\")@

See 'call', 'run', '(->-)', '(=>>-)'.
-}
(=>-) :: IO a           -- ^ Action, whose error output will be redirected
      -> FilePath       -- ^ File to redirect the error output to
      -> IO a           -- ^ Result action
(=>-) =
   redirect_helper stderr WriteMode


{- | Redirect the standard error output of the specified IO action to a file. If the file already exists, the output will be appended.

What's actually modified is the @stderr@ handle, not the file descriptor 2. The
@exec@ functions know about this. See "HsShellScript#fdpipes" and
"HsShellScript#exec" for details.

Example: @call (exec \"\/path\/to\/foo\" [] =\>\>- \"log\")@

See 'call', 'run', '(->>-)', '(=>-)'.
-}
(=>>-) :: IO a          -- ^ Action, whose error output will be redirected
       -> FilePath      -- ^ File to redirect the error output to
       -> IO a          -- ^ Result action
(=>>-) =
   redirect_helper stderr AppendMode


{- | Redirect both stdout and stderr to a file. This is equivalent to the
shell's @&>@ operator. If the file already exists, it will be overwritten.

What's actually modified are the @stdout@ and @stderr@ handles, not the file
descriptors 1 and 2. The @exec@ functions know about this. See
"HsShellScript#fdpipes" and
"HsShellScript#exec" for details.

>(-&>-) io path = err_to_out io ->- path

Example:

@call (exec \"\/path\/to\/foo\" [] -&\>- \"log\")@

See '(-&>>-)', 'err_to_out'.
-}
(-&>-) :: IO a          -- ^ Action, whose output and error output will be redirected
       -> FilePath      -- ^ File to redirect to
       -> IO a          -- ^ Result action
(-&>-) io path = err_to_out io ->- path


{- | Redirect both stdout and stderr to a file. If the file already exists, the
   output will be appended.

What's actually modified are the @stdout@ and @stderr@ handles, not the file
descriptors 1 and 2. The @exec@ functions know about this. See
"HsShellScript#fdpipes" and
"HsShellScript#exec" for details.

   Example:

@call (exec \"\/path\/to\/foo\" [] -&\>\>- \"log\")@

>(-&>>-) io path = (err_to_out >> io) (->>-) path

See '(-&>-)', 'out_to_err'.
-}
(-&>>-) :: IO a         -- ^ Action, whose output and error output will be redirected
       -> FilePath      -- ^ File to redirect to
       -> IO a          -- ^ Result action
(-&>>-) io path =
   err_to_out io ->>- path


{- | Redirect stdin from a file. This modifies the specified action, such
that the standard input is read from a file.

   What's actually modified is the @stdin@ handle, not the file
   descriptor 0. The @exec@ functions know about this. See
   "HsShellScript#fdpipes" and
"HsShellScript#exec" for details.

Example:

@call (exec \"\/path\/to\/foo\" [] -\<- \"bar\")@

See 'exec', 'run', '(->-)', '(=>-)'.
-}
(-<-) :: IO a
      -> FilePath
      -> IO a
(-<-) = redirect_helper stdin ReadMode


-- Helper function for err_to_out and out_to_err
join_helper src tgt io =
   bracket (do bak <- hDuplicate tgt
               hDuplicateTo src tgt
               return bak
           )
           (\bak -> hDuplicateTo bak tgt)
           (const io)


{- | Send the error output of the specified action to its standard output.

What's actually modified is the @stdout@ handle, not the file descriptor 1. The
@exec@ functions know about this. See "HsShellScript#fdpipes" and
"HsShellScript#exec" for details.
-}
err_to_out :: IO a -> IO a
err_to_out = join_helper stdout stderr


{- | Send the output of the specified action to its standard error output.

What's actually modified is the @stderr@ handle, not the file descriptor 2. The
@exec@ functions know about this. See "HsShellScript#fdpipes" and
"HsShellScript#exec" for details.
-}
out_to_err :: IO a -> IO a
out_to_err = join_helper stderr stdout


-- Run an IO action as a new process, and optionally connect its
-- stdin, stdout and stderr via pipes.
pipe_fork_dup :: IO a                   -- Action to run in a new process.
              -> Bool                   -- make stdin pipe?
              -> Bool                   -- make stdout pipe?
              -> Bool                   -- make stderr pipe?
              -> IO ( Maybe Handle      -- Handle to the new process's stdin, if applicable.
                    , Maybe Handle      -- Handle to the new process's stdout, if applicable.
                    , Maybe Handle      -- Handle to the new process's stderr, if applicable.
                    , ProcessID
                    )
pipe_fork_dup io fd0 fd1 fd2 = do
    flush_outerr
    pipe0 <- pipe_if fd0
    pipe1 <- pipe_if fd1
    pipe2 <- pipe_if fd2

    pid <- fmap fromIntegral fork
{-# LINE 553 "src/lib/HsShellScript/ProcErr.chs" #-}
    if pid == 0
       then do
           -- child
           dup_close pipe0 stdin True
           dup_close pipe1 stdout False
           dup_close pipe2 stderr False
           child io
       else do
           -- parent
           h0 <- finish_pipe pipe0 True
           h1 <- finish_pipe pipe1 False
           h2 <- finish_pipe pipe2 False
           return (h0, h1, h2, pid)

  where
     -- Make a pipe, if applicable.
     pipe_if False = return Nothing
     pipe_if True  = fmap Just $ createPipe  -- Just (read,write)

     -- Child work after fork: connect a fd of the new process to the pipe.
     dup_close :: Maybe (Fd, Fd)        -- maybe the pipe
               -> Handle                -- which handle descriptor to connect to the pipe
               -> Bool                  -- whether the child reads from this pipe
               -> IO ()
     dup_close Nothing _ _ =
         return ()
     dup_close m@(Just (readend,writeend)) dest True =
         do h <- System.Posix.fdToHandle readend
            hDuplicateTo h dest
            hClose h
            closeFd writeend
     dup_close m@(Just (readend,writeend)) dest False =
         do h <- System.Posix.fdToHandle writeend
            hDuplicateTo h dest
            hClose h
            closeFd readend

     -- Parent work after fork: close surplus end of the pipe and make a handle from the other end.
     finish_pipe :: Maybe (Fd, Fd)      -- maybe the pipe
                 -> Bool                -- whether the fd is for reading
                 -> IO (Maybe Handle)
     finish_pipe Nothing _ =
         return Nothing
     finish_pipe (Just (readend,writeend)) read =
         do closeFd (if read then readend else writeend)
            let fd = if read then writeend else readend
            h <- System.Posix.fdToHandle fd
            return (Just h)


-- | Run an IO action as a separate process, and pipe some text to its @stdin@.
-- Then close the pipe and wait for the child process to finish. If it
-- exits in a way which indicates an error, the @ProcessStatus@ is thrown as a
-- dynamic exception.
--
-- Example: @pipe_to \"blah\" $ exec \"\/usr\/bin\/foo\" [\"bar\"]@
--
-- See 'call', 'run', '-<-', 'h_pipe_to'. See "HsShellScript#fdpipes" for more details.
pipe_to :: String       -- ^ Text to pipe
        -> IO a         -- ^ Action to run as a separate process, and to pipe to
        -> IO ()
pipe_to str io = do
   (h, pid) <- h_pipe_to io
   hPutStr h str
   hClose h
   (Just ps) <- getProcessStatus True False pid
   if ps == Exited ExitSuccess
       then return ()
       else Control.Exception.throwDyn ps


-- | Run an IO action as a separate process, and connect to its @stdin@
-- with a pipe.
--
-- Example: @h \<- h_pipe_to $ exec \"\/usr\/bin\/foo\" [\"bar\"]@
--
-- See '-<-', 'pipe_to', 'pipe_from', 'pipe_from2'. See "HsShellScript#fdpipes" for more details.
h_pipe_to :: IO a                       -- ^ Action to run as a separate process, and to pipe to
          -> IO (Handle, ProcessID)     -- ^ Returns handle connected to the standard input of the child process, and the child's process ID
h_pipe_to io = do
   (Just h, _, _, pid) <- pipe_fork_dup io True False False
   return (h, pid)


-- | Run an IO action as a separate process, and read its @stdout@
-- strictly. Then wait for the child process to finish. This is like the
-- backquote feature of shells.
--
-- If the child process exits with a non-zero exit code, the
-- @ProcessStatus@ is thrown as a dynamic exception.
--
-- The whole output is returned, no trailing newline character is removed, like the shell does with backquotes. You may want to apply @chomp@
-- to the result.
--
-- Example:
--
-- >output <- pipe_from $ exec "/bin/foo" ["bar"]
--
-- See 'exec', 'pipe_to', 'pipe_from2', 'h_pipe_from', 'lazy_pipe_from', 'chomp', 'silently'. See "HsShellScript#fdpipes" for more details.
pipe_from :: IO a               -- ^ Action to run as a separate process
          -> IO String          -- ^ The called program's standard output
pipe_from io = do
   (h, pid) <- h_pipe_from io
   txt <- hGetContents h
   seq (length txt) (hClose h)
   (Just ps) <- getProcessStatus True False pid
   if ps == Exited ExitSuccess
       then return txt
       else Control.Exception.throwDyn ps


-- | Run an IO action as a separate process, and read its @stderr@
-- strictly. Then wait for the child process to finish, and return the text
-- along with its exit code.
--
-- Example:
--
-- >(errmsg, ec) <- pipe_from2 $ exec "/bin/foo" ["bar"] ->- "/dev/null"
-- >
-- >when (ec /= Exited ExitSuccess) $ do
-- >   errm errmsg
-- >   ...
--
-- See 'exec', 'pipe_to', 'pipe_from', 'h_pipe_from2', 'lazy_pipe_from2', 'silently'. See "HsShellScript#fdpipes" for more details.
pipe_from2 :: IO a                              -- ^ Action to run as a separate process
           -> IO (String, ProcessStatus)        -- ^ The called program's standard output
pipe_from2 io = do
   (h, pid) <- h_pipe_from2 io
   txt <- hGetContents h
   seq (length txt) (hClose h)
   (Just ps) <- getProcessStatus True False pid
   return (txt, ps)


-- | Run an IO action as a separate process, and connect to its @stdout@
-- with a pipe.
--
-- A handle connected to the child process, and the process ID
-- of the child are returned. The process ID can be used with
-- @System.Posix.getProcessStatus@ to get the child's exit code. You must either
-- ensure that all data has been read, or close the handle, before calling
-- @getProcessStatus@ blockingly. Otherwise you'll get a deadlock. When you
-- close the handle before all data has been read, then the child gets a
-- @SIGPIPE@ signal.
--
-- Example:
--
-- >h <- h_pipe_from $ exec "/usr/bin/foo" ["bar"]
--
-- See 'exec', 'pipe_to', 'h_pipe_from2', 'pipe_from', 'lazy_pipe_from', 'chomp', 'silently'. See "HsShellScript#fdpipes" for more details.
h_pipe_from :: IO a                             -- ^ Action to run as a separate process, and to pipe from
            -> IO (Handle, ProcessID)           -- ^ Returns handle connected to the standard output of the child process, and the child's process ID
h_pipe_from io = do
   (_, Just h, _, pid) <- pipe_fork_dup io False True False
   return (h, pid)


-- | Run an IO action as a separate process, and connect to its @stderr@
-- with a pipe.
--
-- A handle connected to the child process' standard error output, and the process ID
-- of the child are returned. The process ID can be used with
-- @System.Posix.getProcessStatus@ to get the child's exit code. You must either
-- ensure that all data has been read, or close the handle, before calling
-- @getProcessStatus@ blockingly. Otherwise you'll get a deadlock. When you
-- close the handle before all data has been read, then the child gets a
-- @SIGPIPE@ signal. Of course, you can also use the process ID to kill the
-- child process.
--
-- Example:
--
-- >h <- h_pipe_from2 $ exec "/usr/bin/foo" ["bar"]
--
-- See 'exec', 'pipe_to', 'h_pipe_from', 'pipe_from2', 'lazy_pipe_from2', 'chomp', 'silently'. See "HsShellScript#fdpipes" for more details.
h_pipe_from2 :: IO a                             -- ^ Action to run as a separate process, and to pipe from
             -> IO (Handle, ProcessID)           -- ^ Returns handle connected to the standard output of the child process, and the child's process ID
h_pipe_from2 io = do
   (_, _, Just h, pid) <- pipe_fork_dup io False False True
   return (h, pid)


-- | Run an IO action as a separate process, and read its @stdout@,
-- This is like the backquote feature of shells. The output is read
-- lazily, as the returned string is evaluated.
--
-- The child's output along with its process ID are returned. The process ID can
-- be used with @System.Posix.getProcessStatus@ to get the child process' exit
-- code. Be aware that you must evaluate the whole string, before calling
-- @getProcessStatus@ blockingly, or you'll get a deadlock. 
--
-- The whole output is returned, no trailing newline character is removed, like
-- the shell does with backquotes. You'll possibly want to apply 'chomp' to the
-- result.
--
-- Example:
--
-- >(txt, pid) <- lazy_pipe_from $ exec "/usr/bin/foo" ["bar"]
-- >...
-- >-- Done, but must read the rest of the output
-- >seq (length txt) (return ())
-- >(Just ps) <- getProcessStatus True False pid
--
-- See 'exec', 'pipe_to', 'pipe_from', 'h_pipe_from', 'lazy_pipe_from2', 'silently'. See "HsShellScript#fdpipes" for more details.
lazy_pipe_from :: IO a                          -- ^ Action to run as a separate process
               -> IO (String, ProcessID)        -- ^ The action's lazy output and the process ID of the child process
lazy_pipe_from io = do
   (_, Just h, _, pid) <- pipe_fork_dup io False True False
   txt <- hGetContents h
   return (txt, pid)


-- | Run an IO action as a separate process, and read its @stderr@. The output
-- is read lazily, as the returned string is evaluated.
--
-- The child's error output along with its process ID are returned. The process
-- ID can be used with @System.Posix.getProcessStatus@ to get the child process'
-- exit code. Be aware that you must evaluate the whole string, before calling
-- @getProcessStatus@ blockingly, or you'll get a deadlock. 
--
-- Example:
--
-- >(errmsg, pid) <- lazy_pipe_from2 $ exec "/usr/bin/foo" ["bar"] ->- "/dev/null"
-- >...
-- >-- Read enough error messages, terminate the child.
-- >signalProcess killProcess pid
-- >
-- >-- Make sure the file descriptor gets closed, or you may run out of file descriptors.
-- >seq (length errmsg) (return ())
--
-- See 'exec', 'pipe_to', 'pipe_from2', 'h_pipe_from2', 'lazy_pipe_from', 'silently'. See "HsShellScript#fdpipes" for more details.
lazy_pipe_from2 :: IO a                          -- ^ Action to run as a separate process
                -> IO (String, ProcessID)        -- ^ The action's lazy output and the process ID of the child process
lazy_pipe_from2 io = do
   (_, Just h, _, pid) <- pipe_fork_dup io False True False
   txt <- hGetContents h
   return (txt, pid)


-- | Run an IO action as a separate process, and optionally connect to its
-- @stdin@, its @stdout@ and its @stderr@ output with
-- pipes.
--
-- See 'pipe_from', 'pipe_from2', 'pipe_to'.
pipes :: IO a                   -- ^ Action to run in a new process
      -> Bool                   -- ^ Whether to make stdin pipe
      -> Bool                   -- ^ Whether to make stdout pipe
      -> Bool                   -- ^ Whether to make stderr pipe
      -> IO ( Maybe Handle
            , Maybe Handle
            , Maybe Handle
            , ProcessID
            )                   -- ^ Pipes to the new process's @stdin@, @stdout@ and @stderr@, if applicable; and its process id.
pipes = pipe_fork_dup


{- | Execute the supplied action. In case of an error, exit with an error
message.

An error is a dynamic exception, thrown using @throwDyn@ as a type which is
instance of @Typeable@. The type err is supposed to be a specific type used
for specific errors. The program is terminated with @exitFailure@.
-}
abort :: Typeable err
      => (err -> String)        -- ^ Error message generation function
      -> IO a                   -- ^ IO action to monitor
      -> IO a                   -- ^ Same action, but abort with error message in case of user exception
abort msgf io =
   catchDyn io
            (\errval -> hPutStrLn stderr (msgf errval) >> exitFailure)


{- | Forcibly terminate the program, circumventing normal program shutdown.

This is the @_exit(2)@ system call. No cleanup actions installed with @bracket@
are performed, no data buffered by file handles is written out, etc.
-}
_exit :: Int                    -- ^ Exit code
      -> IO a                   -- ^ Never returns
_exit ec = do
   _exit_prim (fromIntegral ec)
   throwDyn () -- never reached, only for the type checker


-- | Generate an error message from an @errno@ value. This is the POSIX
-- @strerror@ system library function.
--
-- See the man page @strerror(3)@.
strerror :: Errno       -- ^ @errno@ value
         -> IO String   -- ^ Corresponding error message
strerror (Errno errno) = do
    peekCString (foreign_strerror errno)


-- | Read the global system error number. This is the POSIX @errno@ value. This
-- function is redundant. Use @Foreign.C.Error.getErrno@ instead. 
errno :: IO Errno       -- ^ @errno@ value
errno = getErrno


-- | Print error message corresponding to the specified @errno@ error
-- number. This is similar to the POSIX system library function @perror@.
--
-- See the man page @perror(3)@.
perror' :: Errno        -- ^ @errno@ error number
        -> String       -- ^ Text to precede the message, separated by \"@: @\"
        -> IO ()
perror' errno txt = do
   str <- strerror errno
   hPutStrLn stderr ((if txt == "" then "" else txt ++ ": ") ++ str)


-- | Print error message corresponding to the global @errno@ error
-- number. This is the same as the POSIX system library function @perror@.
--
-- See the man page @perror(3)@.
perror :: String        -- ^ Text to precede the message, separated by \"@: @\"
       -> IO ()
perror txt = do
   eno <- getErrno
   perror' eno txt


-- | Print a message to @stderr@ and exit with an exit code
-- indicating an error.
--
-- >failIO msg = hPutStrLn stderr msg >> exitFailure
failIO :: String -> IO a
failIO meld =
   hPutStrLn stderr meld >> exitFailure


-- |Return the exit code, instead of throwing it as a dynamic exception.
--
-- This is used to modify the error reporting behaviour of an IO action. It
-- is used in conjunction with 'run' or 'call'. Normally, they throw a
-- process status, which indicates any error as a dynamic exception. After
-- @exitcode@ has been applied, only termination by a signal causes an exception
-- to be thrown. Any exit code is returned instead.
--
-- Example: @ec \<- exitcode $ run \"foo\" [\"bar\"]@
--
-- See 'run', 'call'.
exitcode :: IO ()             -- ^ Action to modify
         -> IO ExitCode       -- ^ Modified action
exitcode io =
   do io
      return ExitSuccess
   `Control.Exception.catchDyn`
      (\processstatus ->
          case processstatus of
             (Exited ec) -> return ec
             ps          -> Control.Exception.throwDyn ps)


-- |Create and throw an @IOError@ from the current @errno@ value, an optional handle and an optional file name.
--
-- This is an extended version of the @Foreign.C.Error.throwErrno@ function
-- from the GHC libraries, which additionally allows to specify a handle and a file
-- name to include in the @IOError@ thrown.
--
-- See @Foreign.C.Error.throwErrno@, @Foreign.C.Error.errnoToIOError@.
throwErrno' :: String           -- ^ Description of the location where the error occurs in the program
            -> Maybe Handle     -- ^ Optional handle
            -> Maybe FilePath   -- ^ Optional file name (for failing operations on files)
            -> IO a
throwErrno' loc maybe_handle maybe_filename =
  do
    errno <- getErrno
    ioError (errnoToIOError loc errno maybe_handle maybe_filename)


-- |Convert an @IOError@ to a string.
--
-- There is an instance declaration of @IOError@ in @Show@ in the @GHC.IOBase@ library, but @show_ioerror@ produces a more readable, and more
-- complete, message.
show_ioerror :: IOError -> String
show_ioerror ioe =
   "IO-Error\n\
   \   Error type:   " ++ show (ioe_type ioe) ++ "\n\
   \   Location:     " ++ none (indent (ioe_location ioe)) ++ "\n\
   \   Description:  " ++ none (indent (ioe_description ioe)) ++ "\n\
   \   " ++ fn (ioeGetFileName ioe)
   where fn (Just n) = "File name:    " ++ quote n
         fn Nothing  = "File name:    (none)"
         none ""  = "(none)"
         none msg = msg
         indent txt = concat (intersperse ("\n                 ") (lines txt))


{- |
   Call the shell to execute a command. In case of an error, throw the @ProcessStatus@ (such as @(Exited (ExitFailure ec))@) as a dynamic exception.
   This is like the Haskell standard library function @system@, except that error handling is brought in accordance with HsShellScript's scheme.

   @exitcode . system_throw@ is the same as the @system@ function, except that when the called shell is terminated or stopped by a signal, this still
   leads to the @ProcessStatus@ thrown as a dynamic exception. The Haskell library report says nothing about what happens in this case, when using the
   @system@ function.

>system_throw cmd = run "/bin/sh" ["-c", "--", cmd]
-}
-- This function should go to HsShellScript.Shell, but this would introduce a circular dependency.
system_throw :: String -> IO ()
system_throw cmd =
   run "/bin/sh" ["-c", "--", cmd]


{- | Run a subroutine as a child process, but don't let it produce any messages.
Read its @stdout@ and @stderr@ instead, and append it to the contents of a
mutable variable. The idea is that you can run some commands silently, and
report them and their messages to the user only when something goes wrong.

If the child process terminates in a way which indicates an error, then the
process status is thrown, in the same way as 'run' does. If the subroutine
throws an @(Exited ec)@ exception (of type @ProcessStatus@), such as thrown by
'run', then the child process exits with the same exit code, such that the
parent process reports it to the caller, again as a @ProcessStatus@ exception.

When the subroutine finishes, the child process is terminated with @'_exit' 0@.
When it throws an exception, an error message is printed and it is terminated
with @'_exit' 1@. See "HsShellScript#subr" for details.

The standard output (and the standard error output) of the parent process are
flushed before the fork, such that no output appears twice.

Example:

>let handler :: IORef String -> ProcessStatus -> IO ()
>    handler msgref ps = do hPutStrLn stderr ("Command failed with " ++ show ps ++ ". Actions so far: ")
>                           msg <- readIORef msgref
>                           hPutStrLn stderr msg
>                           exitWith (ExitFailure 1)
>
>msgref <- newIORef ""
>do silently msgref $ do putStrLn "Now doing foobar:"
>                        echo exec "/foo/bar" ["arguments"]
>   silently msgref $ echo exec "/bar/baz" ["arguments"]
>`catchDyn` (handler msgref)

See 'lazy_pipe_from', 'call', 'run', GHC.IOBase.IORef.
-}
silently :: GHC.IOBase.IORef String             -- ^ A mutable variable, which gets the output (stdout and stderr) of the action appended.
         -> IO ()                               -- ^ The IO action to run.
         -> IO ()
silently ref io = do
   (msg, pid) <- lazy_pipe_from (err_to_out (child io))
   seq (length msg) (return ())

   msgs <- readIORef ref
   writeIORef ref (msgs ++ msg)

   (Just ps) <- getProcessStatus True False pid
   case ps of
      Exited ExitSuccess -> return ()
      ps                 -> Control.Exception.throwDyn ps


{- | Modify a subroutine action in order to make it suitable to run as a child
   process.

   This is used by 'call', 'silently', 'pipe_to', 'exec' etc. The action is
   executed. When it returns, the (child) process is terminated with @'_exit' 0@
   (after flushing @stdout@), circumventing normal program shutdown. When it
   throws an exception, an error message is printed and the (child) process is
   terminated with @'_exit' 1@.
-}
child :: IO a           -- Action to modify
      -> IO b           -- Never returns
child io = do
   (io `finally` flush_outerr)
      `catchDyn`
         (\argerror -> do
             errm $ "In child process:\n" ++ argerror_message argerror
             _exit 1
         )
      `catchDyn`
         (\processstatus -> do
             errm $ "Process error in child process. Process status = " ++ show ( processstatus :: ProcessStatus )
             _exit 1
         )
      `IO.catch`
         (\ioe -> do
             errm ("In child process:\n" ++ show_ioerror ioe)
             _exit 1
         )
      `Control.Exception.catch`
          (\e -> case e of
                    ExitException ec ->
                       do -- Child process is a subroutine that has terminated normally.
                          errm "Warnung! Child process tries to shut down normally. This is a bug. It should\n\
                               \terminate with _exit (or catch the ExitException yourself). See section\n\"\
                               \Running a Subroutine in a Separate Process\" in the HsShellScript API\n\
                               \documentation. Terminating with _exit 0 now."
                          _exit (case ec of
                                    ExitSuccess    -> 0
                                    ExitFailure ec -> ec
                                )
                    _ -> do errm $ "Child process quit with unexpected exception:\n" ++ show e
                            _exit 1
          )
   _exit 0



{- | Colorful log message to @stderr@.

   This prints a message to @stderr@. When @stderr@ is connected to a terminal
   (as determined by @isatty(3)@), additional escape sequences are printed,
   which make the message appear in cyan.

   Additionally, a newline character is output at the end.

   See 'logm_', 'errm', 'errm_'.
-}
logm :: String          -- ^ Message to print
     -> IO ()
logm msg =
   do fd <- get_fd stderr
      isterm <- isatty fd
      if (isterm /= 0)
         then hPutStrLn stderr $ "\ESC[36m" ++ msg ++ "\ESC[00m"
         else hPutStrLn stderr msg


{- | Colorful log message to @stderr@.

   This prints a message to @stderr@. When @stderr@ is connected to a terminal
   (as determined by @isatty(3)@), additional escape sequences are printed,
   which make the message appear in cyan.

   No a newline character is output at the end.

   See 'logm', 'errm', 'errm_'.
-}
logm_ :: String -> IO ()
logm_ msg = do
   do fd <- get_fd stderr
      isterm <- isatty fd
      if (isterm /= 0)
         then hPutStr stderr $ "\ESC[36m" ++ msg ++ "\ESC[00m"
         else hPutStr stderr msg


{- | Colorful error message to @stderr@.

   This prints a message to @stderr@. When @stderr@ is connected to a terminal
   (as determined by @isatty(3)@), additional escape sequences are printed,
   which make the message appear in red.

   Additionally, a newline character is output at the end.

   See 'logm', 'logm_', 'errm_'.
-}
errm :: String -> IO ()
errm msg = do
   do fd <- get_fd stderr
      isterm <- isatty fd
      if (isterm /= 0)
         then hPutStrLn stderr $ "\ESC[01;31m" ++ msg ++ "\ESC[00m"
         else hPutStrLn stderr msg


{- | Colorful error message to @stderr@.

   This prints a message to @stderr@. When @stderr@ is connected to a terminal
   (as determined by @isatty(3)@), additional escape sequences are printed,
   which make the message appear in red.

   No a newline character is output at the end.

   See 'logm', 'logm_', 'errm'.
-}
errm_ :: String -> IO ()
errm_ msg = do
   do fd <- get_fd stderr
      isterm <- isatty fd
      if (isterm /= 0)
         then hPutStr stderr $ "\ESC[01;31m" ++ msg ++ "\ESC[00m"
         else hPutStr stderr msg


{- | In case the specified action throws an IOError, fill in its filename field. This way, more useful error messages can be produced.

Example:

>-- Oh, the GHC libraries neglect to fill in the file name
>executeFile' prog a b c =
>   fill_in_filename prog $ executeFile prog a b c

See 'fill_in_location', 'add_location'.
-}
fill_in_filename :: String              -- ^ File name to fill in
                 -> IO a                -- ^ IO action to modify
                 -> IO a                -- ^ Modified IO action
fill_in_filename filename io =
   io `IO.catch` (\ioe -> ioError (ioe { ioe_filename = Just filename }))


{- | In case the specified action throws an IOError, fill in its location field. This way, more useful error messages can be produced.

Example:

>my_fun a b c = do
>   -- ...
>   fill_in_location "my_fun" $  -- Give the caller a more useful location information in case of failure
>      rename "foo" "bar"
>   -- ...

See 'fill_in_filename'.
-}
fill_in_location :: String              -- ^ Location name to fill in
                 -> IO a                -- ^ IO action to modify
                 -> IO a                -- ^ Modified IO action
fill_in_location location io =
   io `IO.catch` (\ioe -> ioError (ioe { ioe_location = location }))


{- | In case the specified action throws an IOError, add a line to its location field. This way, more useful error messages can be produced. The
   specified string is prepended to the old location, separating it with a newline from the previous location, if any. When using this thoroughly, you
   get a reverse call stack in IOErrors.

Example:

>my_fun =
>   add_location "my_fun" $ do
>      -- ...

See 'fill_in_filename', 'fill_in_location'.
-}
add_location :: String              -- ^ Location name to add
             -> IO a                -- ^ IO action to modify
             -> IO a                -- ^ Modified IO action
add_location location io =
   io `IO.catch` (\ioe -> let loc = case ioe_location ioe of
                                       ""   -> location
                                       loc0 -> location ++ "\n" ++ loc0
                          in  ioError (ioe { ioe_location = loc })
                 )


{- | This is a replacement for @System.Posix.Process.executeFile@. It does
   additional preparations, then calls @executeFile@.
   @executeFile@ /can't normally be used directly, because it doesn't do the
   things which are outlined here./

   This are the differences to @executeFile@:

   1. @stdout@ and @stderr@ are flushed.

   2. The standard file descriptors 0-2 are made copies of the file descriptors
   which the standard handles currently use. This is necessary because they
   might no longer use the standard handles. See "HsShellScript#fdpipes".

   If the standard handles @stdin@, @stdout@, @stderr@ aren't in closed state,
   and they aren't already connected to the respective standard file
   descriptors, their file descriptors are copied to the respective standard
   file descriptors (with @dup2@). Backup copies are made of the file
   descriptors which are overwritten. If some of the standard handles are closed,
   the corresponding standard file descriptors are closed as well.

   3. All file descriptors, except for the standard ones, are set to close on
   exec (see @fcntl(2)@), and will be closed on successful replacement of
   the process. Before that, the old file descriptor flags are saved.

   4. The standard file descriptors are set to blocking mode, since GHC 6.2.2
   sets file descriptors to non-blocking (except 0-2, which may get 
   overwritten by a non-blocking one in step 2). The called program
   doesn't expect that.

   5. In case replacing the process fails, the file descriptors are reset to
   the original state. The file descriptors flags are restored, and the file
   descriptors 0-2 are overwritten again, with their backup copies. Then an
   IOError is thrown. 

   6. In any IOError, the program is filled in as the file name (@executeFile@
   neglects this).

   7. The return type is a generic @a@, rather than @()@.

   Also see "HsShellScript#exec".
-}
execute_file :: FilePath                     -- ^ Program to call
             -> Bool                         -- ^ Search @PATH@?
             -> [String]                     -- ^ Arguments
             -> Maybe [(String, String)]     -- ^ Optionally new environment
             -> IO a                         -- ^ Never returns
execute_file path search args menv =
   fill_in_filename path $ fill_in_location "execute_file" $ 
      bracket (do -- Flush stdout and stderr, if open
                  flush_outerr

                  -- Make fds 0-2 copies of the things which the standard handles refer to.
                  recover0 <- restore stdin 0
                  recover1 <- restore stdout 1
                  recover2 <- restore stderr 2

                  -- Save the flags of all file descriptors
                  fdflags <- c_save_fdflags
{-# LINE 1242 "src/lib/HsShellScript/ProcErr.chs" #-}

                  -- Prepare all fds for subsequent exec. Fds 0-2 are set to blocking (because GHC sets new fds to non-blocking). All others are set
                  -- to close-on-exec.
                  c_prepare_fd_flags_for_exec
{-# LINE 1246 "src/lib/HsShellScript/ProcErr.chs" #-}

                  return (recover0, recover1, recover2, fdflags)
               )
               (\(recover0, recover1, recover2, fdflags) ->
                   do -- Failure of the exec. Restore the file descriptor flags
                      c_restore_fdflags fdflags

                      -- Restore the standard handles
                      recover0
                      recover1
                      recover2
               )
               (const $ do 
                   -- The exec
                   executeFile path search args menv

                   -- Never reached, only for the type checker
                   throwDyn ()
               )

   where
      restore h@(FileHandle _ mvar) fd = do
         -- Get the fd which the handle h uses. This locks the handle.
         h__ <- takeMVar mvar

         -- Make a copy of the fd which is about to be overwritten. Returns -1 for invalid (closed) fd.
         fd_backup <- c_fcntl_dupfd fd 3

         let -- The fd used by the handle. This is in GHC.IOBase.
             handle_fd = fromIntegral (haFD h__)

             -- Is the handle closed?
             closed = case haType h__ of
                         ClosedHandle -> True
                         SemiClosedHandle -> True
                         _ -> False

         -- If the handle is open, make the fd a copy of the fd which the handle uses. Otherwise, close the fd as well.
         if closed
            then close fd >> return ()
            else when (fd /= handle_fd) $ dup2 handle_fd fd >> return ()

         -- Return recovery action which undoes everything.
         return (do -- Restore the fd
                    if fd_backup /= -1 then do dup2 fd_backup fd
                                               close fd_backup
                                               return ()
                                       else do close fd
                                               return ()
                    -- Unlock the handle
                    putMVar mvar h__
                    return ()
                )

      -- Silly: The standard handle has been overwritten with a duplex.
      restore h fd = do
         -- Make a copy of the fd which is about to be closed. Returns -1 for already closed fd.
         fd_backup <- c_fcntl_dupfd fd 3

         -- Close the fd
         close fd

         -- Return recovery action, which restores the fd.
         return (if fd_backup /= -1 then do dup2 fd_backup fd
                                            close fd_backup
                                            return ()
                                    else do close fd
                                            return ()
                )



-- Get the file descriptor which is encapsulated in a file handle.
get_fd h@(FileHandle _ mvar) = do
   h__ <- takeMVar mvar
   let fd = fromIntegral (haFD h__)
   putMVar mvar h__
   return fd
get_fd _ =
   -- Duplex 
   return (-1)


-- Flush stdout and stderr (which should not be necessary). Discard Illegal Operation IOError which arises when they are closed.
flush_outerr = do
   flush stdout
   flush stderr
   where flush h = hFlush h `Prelude.catch` (\ioe -> if isIllegalOperation ioe then return () else ioError ioe)


{-
data ProcessStatus = Exited ExitCode
                   | Terminated Signal
                   | Stopped Signal
		   deriving (Eq, Ord, Show)
-}
instance Typeable ProcessStatus where
   typeOf = const tyCon_ProcessStatus

tyCon_ProcessStatus = mkAppTy (mkTyCon "Posix.ProcessStatus") []


-- | The GHC libraries don't declare @Foreign.C.Error.Errno@ as instance of
-- Show. This makes it up.
instance Show Foreign.C.Error.Errno where
   show (Errno e) = show e

foreign import ccall safe "build/HsShellScript/ProcErr.h fork"
  fork :: (IO CInt)

foreign import ccall safe "build/HsShellScript/ProcErr.h _exit"
  _exit_prim :: (CInt -> (IO ()))

foreign import ccall safe "build/HsShellScript/ProcErr.h strerror"
  foreign_strerror :: (CInt -> (Ptr CChar))

foreign import ccall safe "build/HsShellScript/ProcErr.h isatty"
  isatty :: (CInt -> (IO CInt))

foreign import ccall safe "build/HsShellScript/ProcErr.h c_save_fdflags"
  c_save_fdflags :: (IO (Ptr CInt))

foreign import ccall safe "build/HsShellScript/ProcErr.h c_prepare_fd_flags_for_exec"
  c_prepare_fd_flags_for_exec :: (IO ())

foreign import ccall safe "build/HsShellScript/ProcErr.h c_restore_fdflags"
  c_restore_fdflags :: ((Ptr CInt) -> (IO ()))

foreign import ccall safe "build/HsShellScript/ProcErr.h c_fcntl_dupfd"
  c_fcntl_dupfd :: (CInt -> (CInt -> (IO CInt)))

foreign import ccall safe "build/HsShellScript/ProcErr.h close"
  close :: (CInt -> (IO CInt))

foreign import ccall safe "build/HsShellScript/ProcErr.h dup2"
  dup2 :: (CInt -> (CInt -> (IO CInt)))
